// (C) 2001-2021 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


// synopsys translate_off
`timescale 1 ns / 1 ps
// synopsys translate_on

module intel_pcie_dma_hwtcl # (
      parameter PFNUM_WIDTH 				                            = 2,
      parameter VFNUM_WIDTH 				                            = 11,
      parameter DESC_DATA_WIDTH                                         = 18,
      parameter DESC_BE_WIDTH                                           = DESC_DATA_WIDTH/32,
      parameter STORAGE_FIFO_DEPTH                                      = 7
   ) 
   (
        input   logic                          	                        Clk_i,
        input   logic                          	                        Rstn_i,
        
        // DMA Controller Avalon-MM slave (from BRI)
        input   logic [11:0] 	                                        dma_s_address_i,
        input   logic								                    dma_s_read_i,
        output  logic [511:0]					                        dma_s_readdata_o,
        output  logic								                    dma_s_readdatavalid_o,
        input   logic								                    dma_s_write_i,
        input   logic [511:0]					                        dma_s_writedata_i, 
        output  logic								                    dma_s_waitrequest_o,
        input   logic [63:0]                                            dma_s_byteenable_i,
        output  logic [1:0]                                             dma_s_response_o,
        input   logic [3:0]                                             dma_s_burstcount_i,

        // DMA Controller Interface
        // Write data mover interface
        output  logic                                                   dma_wrdm_desc_valid_o,
        output  logic [1+VFNUM_WIDTH+PFNUM_WIDTH+159:0]                 dma_wrdm_desc_data_o,
        input   logic                                                   dma_wrdm_desc_ready_i,

        output  logic                                                   dma_wrdm_prio_valid_o,
        output  logic [1+VFNUM_WIDTH+PFNUM_WIDTH+159:0]                 dma_wrdm_prio_data_o,
        input   logic                                                   dma_wrdm_prio_ready_i,

        input   logic [1+VFNUM_WIDTH+PFNUM_WIDTH+DESC_DATA_WIDTH-1:0]   dma_wrdm_tx_data_i,
        input   logic                                                   dma_wrdm_tx_valid_i,

        // Read data mover interface
        output  logic                                                   dma_rddm_desc_valid_o,
        output  logic [1+VFNUM_WIDTH+PFNUM_WIDTH+159:0]                 dma_rddm_desc_data_o,
        input   logic                                                   dma_rddm_desc_ready_i,

        output  logic                                                   dma_rddm_prio_valid_o,
        output  logic [1+VFNUM_WIDTH+PFNUM_WIDTH+159:0]                 dma_rddm_prio_data_o,
        input   logic                                                   dma_rddm_prio_ready_i,

        input   logic [1+VFNUM_WIDTH+PFNUM_WIDTH+DESC_DATA_WIDTH-1:0]   dma_rddm_tx_data_i,
        input   logic                                                   dma_rddm_tx_valid_i,
        
        //Dummy pfnum ports
        input   wire  [PFNUM_WIDTH-1:0]                                 dma_wrdm_pfnum_i,
        input   wire  [PFNUM_WIDTH-1:0]                                 dma_rddm_pfnum_i, 

// Avalon-MM slave (from BRI)
      input  logic   [14:0]  	msix_table_address_i,       
      input  logic   [63:0]   	msix_table_byteenable_i,    
      input  logic           	msix_table_read_i,          
      output logic   [511:0] 	msix_table_readdata_o,
	  output logic   			msix_table_readdatavalid_o,	
      input  logic           	msix_table_write_i,         
      input  logic   [511:0]  	msix_table_writedata_i,     
      input logic [3:0] 		msix_table_burstcount_i,
	  output logic           	msix_table_waitrequest_o 
	
   );

   
logic [3:0]                               int_storage_reg[3:0];
logic [3:0]                               int_storage_wr;
logic [4:0]                               int_storage_rd;
logic [7:0]                               int_desc_id;
logic [2:0]                               int_storage_info;
logic [3:0]                               int_storage_info_wire;
logic [4:0]                               fill_int_info;

logic                                     wr_int_info;
logic                                     wr_int_info_rddm;
logic                                     wr_int_info_wrdm;
logic                                     read_int_info;
logic                                     empty_int_storage;
logic                                     fill_int_less_than4;
logic [2:0]                               int_app_spec;
logic [DESC_DATA_WIDTH-1:0]               wrdm_int_data;
logic [DESC_DATA_WIDTH-1:0]               rddm_int_data;
logic [DESC_DATA_WIDTH-1:0]               int_desc_wire;
logic [DESC_DATA_WIDTH*2-1:0]             wrdm_prio_data;
logic [DESC_BE_WIDTH*2-1:0]               wrdm_prio_be;
logic                                     wrdm_prio_write;
logic                                     wen_wrdm_desc;
logic                                     wen_wrdm_prio;
logic                                     wen_rddm_prio;
logic                                     wen_rddm_desc;
logic                                     wen_wrdm_int;
logic                                     wen_rddm_int;
logic                                     dma_wrdm_desc_full_o;
logic                                     dma_wrdm_prio_full_o;
logic                                     dma_rddm_desc_full_o;
logic                                     dma_rddm_prio_full_o;
logic [DESC_DATA_WIDTH*2-1:0]             descriptor_data_r,descriptor_data_wire;
logic [DESC_BE_WIDTH * 2-1:0]             descriptor_be_wire;
logic [DESC_BE_WIDTH * 2-1:0]             be_wrdm_desc;
logic [DESC_BE_WIDTH * 2-1:0]             be_wrdm_prio;
logic [DESC_BE_WIDTH * 2-1:0]             be_rddm_prio;
logic [DESC_BE_WIDTH * 2-1:0]             be_rddm_desc;


logic [511:0]                             descriptor_int_data_r;
logic [511:0]                             wrdm_int_reg;
logic [511:0]                             rddm_int_reg;
logic [7:0]                               fillcount_wrdm_prio;
logic [7:0]                               fillcount_wrdm_desc;
logic [7:0]                               fillcount_rddm_prio;
logic [7:0]                               fillcount_rddm_desc;
logic                                     read_prio_queue_status;
logic                                     read_wrdm_queue_status,read_wrdm_queue_status_r;
logic                                     read_w_status         ,read_w_status_r;
logic                                     read_r_status         ,read_r_status_r;
logic                                     read_wrdm_int, read_rddm_int, read_wrdm_int_r, read_rddm_int_r ; 
logic [31:0]                              read_data_wrdm;
logic [31:0]                              read_data_rddm;
logic [31:0]                              read_rddm_status;
logic [31:0]                              read_wrdm_status;
logic [511:0]                             read_data_final;
logic [31:0]                              wrdm_status_queue_rdata;
logic [31:0]                              rddm_status_queue_rdata;
logic                                     read_data_valid;
logic                                     dma_s_read_r;
logic                                     dma_s_read_rr;
logic                                     wrdm_status_queue_empty;
logic                                     rddm_status_queue_empty;
logic [1:0]                               int_info_wrdm,int_info_rddm;

assign dma_s_response_o = 2'b0;

  assign wrdm_prio_data  = wen_wrdm_prio ? descriptor_data_r : read_int_info ? {{160{1'bX}},int_desc_wire} : {{320{1'bX}}};
  assign wrdm_prio_be    = wen_wrdm_prio ? be_wrdm_prio : read_int_info ? {5'h00,5'h1F} : 10'h000;
  assign wrdm_prio_write = read_int_info | wen_wrdm_prio;
  
  assign true_write = !dma_s_waitrequest_o & dma_s_write_i;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////// Address arbiter for the write side of the slave logic.
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      wen_wrdm_desc <= '0;
      wen_wrdm_prio <= '0;
      wen_rddm_prio <= '0;
      wen_rddm_desc <= '0;
      wen_wrdm_int  <= '0;
      wen_rddm_int  <= '0;
      be_wrdm_desc  <= '0;
      be_wrdm_prio  <= '0;
      be_rddm_prio  <= '0;
      be_rddm_desc  <= '0;
    end
    else begin
      wen_wrdm_desc <= (dma_s_address_i[11:9] == 3'h0) ? true_write : '0;
      wen_wrdm_prio <= (dma_s_address_i[11:9] == 3'h1) ? true_write : '0;
      wen_rddm_desc <= (dma_s_address_i[11:9] == 3'h4) ? true_write : '0;
      wen_rddm_prio <= (dma_s_address_i[11:9] == 3'h5) ? true_write : '0;
      wen_wrdm_int  <= (dma_s_address_i[11:9] == 3'h3) ? true_write : '0;
      wen_rddm_int  <= (dma_s_address_i[11:9] == 3'h7) ? true_write : '0;
      
      be_wrdm_desc  <= ((dma_s_address_i[11:9] == 3'h0) & true_write) ? descriptor_be_wire : '0;
      be_wrdm_prio  <= ((dma_s_address_i[11:9] == 3'h1) & true_write) ? descriptor_be_wire : '0;
      be_rddm_desc  <= ((dma_s_address_i[11:9] == 3'h4) & true_write) ? descriptor_be_wire : '0;
      be_rddm_prio  <= ((dma_s_address_i[11:9] == 3'h5) & true_write) ? descriptor_be_wire : '0;
    end
   end

assign dma_s_waitrequest_o = dma_wrdm_desc_full_o | dma_wrdm_prio_full_o | dma_rddm_desc_full_o | dma_rddm_prio_full_o;


  always_ff @(posedge Clk_i) begin
    descriptor_data_r     <= descriptor_data_wire;
    //descriptor_int_data_r <= dma_s_writedata_i;  
  end

  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
       wrdm_int_reg <= '0;
       rddm_int_reg <= '0;
    end
    else begin
       wrdm_int_reg <= wen_wrdm_int ? descriptor_int_data_r : wrdm_int_reg;
       rddm_int_reg <= wen_rddm_int ? descriptor_int_data_r : rddm_int_reg;
    end
  end
  
  genvar unmasked_byte;
  generate  
      for (unmasked_byte = 0; unmasked_byte < 64; unmasked_byte = unmasked_byte + 1)
        begin
          always @ (posedge Clk_i)
            if (dma_s_byteenable_i[unmasked_byte] == 1)
            descriptor_int_data_r[(unmasked_byte*8)+7:(unmasked_byte*8)] <=  dma_s_writedata_i[(unmasked_byte*8)+7:(unmasked_byte*8)];       

        end
  endgenerate

  assign descriptor_data_wire = {dma_s_writedata_i[DESC_DATA_WIDTH+255:256],dma_s_writedata_i[DESC_DATA_WIDTH-1:0]};

  genvar i;
  generate 
    for (i=0; i < (DESC_BE_WIDTH * 2); i=i+1 ) begin : be_select
      if (i > 4)
        assign descriptor_be_wire[i] = dma_s_byteenable_i[i*4 + 15];
      else
        assign descriptor_be_wire[i] = dma_s_byteenable_i[i*4 + 3];
    end
  endgenerate    

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////// Address and data arbiter for the read side of the slave logic.
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
       fillcount_wrdm_prio <= '0;
       fillcount_wrdm_desc <= '0;
       fillcount_rddm_prio <= '0;
       fillcount_rddm_desc <= '0;
    end
    else begin
      if (dma_wrdm_tx_valid_i & !dma_wrdm_desc_valid_o & !dma_wrdm_tx_data_i[8])
         fillcount_wrdm_desc <= fillcount_wrdm_desc - 1'b1;      
      if (!dma_wrdm_tx_valid_i & dma_wrdm_desc_valid_o)
         fillcount_wrdm_desc <= fillcount_wrdm_desc + 1'b1;
      
      if (dma_wrdm_tx_valid_i & !dma_wrdm_prio_valid_o & dma_wrdm_tx_data_i[8])
         fillcount_wrdm_prio <= fillcount_wrdm_prio - 1'b1;
      if (!dma_wrdm_tx_valid_i & dma_wrdm_prio_valid_o)
         fillcount_wrdm_prio <= fillcount_wrdm_prio + 1'b1;
         
      if (dma_rddm_tx_valid_i & !dma_rddm_desc_valid_o & !dma_rddm_tx_data_i[8])
         fillcount_rddm_desc <= fillcount_rddm_desc - 1'b1;      
      if (!dma_rddm_tx_valid_i & dma_rddm_desc_valid_o)
         fillcount_rddm_desc <= fillcount_rddm_desc + 1'b1;
         
      if (dma_rddm_tx_valid_i & !dma_rddm_prio_valid_o & dma_rddm_tx_data_i[8])
         fillcount_rddm_prio <= fillcount_rddm_prio - 1'b1;      
      if (!dma_rddm_tx_valid_i & dma_rddm_prio_valid_o)
         fillcount_rddm_prio <= fillcount_rddm_prio + 1'b1;

    end
  end

  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      read_prio_queue_status <= '0;
      read_wrdm_queue_status <= '0;
      read_w_status          <= '0;
      read_r_status          <= '0;
      read_wrdm_int          <= '0;     
      read_rddm_int          <= '0;
    end
    else begin
      read_prio_queue_status <= dma_s_read_i & (dma_s_address_i[10:9] == 2'h1);
      read_wrdm_queue_status <= dma_s_read_i & (dma_s_address_i[11:10] == 2'h0);
      read_w_status          <= dma_s_read_i & (dma_s_address_i[11:9] == 3'h2);
      read_r_status          <= dma_s_read_i & (dma_s_address_i[11:9] == 3'h6);
      read_wrdm_int          <= dma_s_read_i & (dma_s_address_i[11:9] == 3'h3);
      read_rddm_int          <= dma_s_read_i & (dma_s_address_i[11:9] == 3'h7);
    end
  end
  
  always_ff @(posedge Clk_i) begin
    read_wrdm_queue_status_r <= read_wrdm_queue_status;
    read_w_status_r          <= read_w_status;
    read_r_status_r          <= read_r_status;
    read_wrdm_int_r          <= read_wrdm_int; 
    read_rddm_int_r          <= read_rddm_int;
    dma_s_read_r             <= dma_s_read_i;
    dma_s_read_rr            <= dma_s_read_r;
    read_data_wrdm           <= read_prio_queue_status ? {dma_wrdm_prio_ready_i,23'h000000,fillcount_wrdm_prio} : {dma_wrdm_desc_ready_i,23'h000000,fillcount_wrdm_desc};
    read_data_rddm           <= read_prio_queue_status ? {dma_rddm_prio_ready_i,23'h000000,fillcount_rddm_prio} : {dma_rddm_desc_ready_i,23'h000000,fillcount_rddm_desc};    
    read_wrdm_status         <= wrdm_status_queue_empty ? {wrdm_status_queue_empty,31'h00000000}  : wrdm_status_queue_rdata;
    read_rddm_status         <= rddm_status_queue_empty ? {rddm_status_queue_empty,31'h00000000}  : rddm_status_queue_rdata;
    read_data_final          <= read_rddm_int_r ? rddm_int_reg : read_wrdm_int_r ? wrdm_int_reg : read_w_status_r ? read_wrdm_status : read_r_status_r ? read_rddm_status : read_wrdm_queue_status_r ? read_data_wrdm : read_data_rddm;
    read_data_valid          <= dma_s_read_rr;
    //read_data_valid          <= read_rddm_int_r | read_wrdm_int_r | read_w_status_r | read_r_status_r | read_wrdm_queue_status_r;
  end
  
  assign dma_s_readdatavalid_o = read_data_valid;
  assign dma_s_readdata_o      = read_data_final;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



// Fifos for the descriptors of the individual movers
intel_pcie_descriptor_storage_fifo #(
  
  .DESC_DATA_WIDTH     (DESC_DATA_WIDTH),
  .PFNUM_WIDTH         (PFNUM_WIDTH),
  .VFNUM_WIDTH         (VFNUM_WIDTH),
  .DESC_BE_WIDTH       (DESC_DATA_WIDTH/32),
  .STORAGE_FIFO_DEPTH  (STORAGE_FIFO_DEPTH)

) write_desc_queue
(
    .clk                     (Clk_i),
    .rst_n                   (Rstn_i),
    
    .dma_s_wdata_i           (descriptor_data_r),
    .dma_s_be_i              (be_wrdm_desc),
    .dma_s_write_i           (wen_wrdm_desc),    
    .ready_i                 (dma_wrdm_desc_ready_i),    
    .desc_rdata_o            (dma_wrdm_desc_data_o),
    .data_valid_o            (dma_wrdm_desc_valid_o),
    .full_o                  (dma_wrdm_desc_full_o),
    .fillcount_o             ()


);

intel_pcie_descriptor_storage_fifo #(

  .DESC_DATA_WIDTH (DESC_DATA_WIDTH),
  .PFNUM_WIDTH  (PFNUM_WIDTH),
  .VFNUM_WIDTH  (VFNUM_WIDTH),
  .DESC_BE_WIDTH  (DESC_DATA_WIDTH/32),
  .STORAGE_FIFO_DEPTH  (STORAGE_FIFO_DEPTH)

) write_prio_queue
(
    .clk                     (Clk_i),
    .rst_n                   (Rstn_i),
    
    .dma_s_wdata_i           (wrdm_prio_data),
    .dma_s_be_i              (wrdm_prio_be),
    .dma_s_write_i           (wrdm_prio_write),
    .ready_i                 (dma_wrdm_prio_ready_i),    
    .desc_rdata_o            (dma_wrdm_prio_data_o),
    .data_valid_o            (dma_wrdm_prio_valid_o),
    .full_o                  (dma_wrdm_prio_full_o),
    .fillcount_o             ()


);

intel_pcie_descriptor_storage_fifo #(
  
  .DESC_DATA_WIDTH (DESC_DATA_WIDTH),
  .PFNUM_WIDTH  (PFNUM_WIDTH),
  .VFNUM_WIDTH  (VFNUM_WIDTH),
  .DESC_BE_WIDTH  (DESC_DATA_WIDTH/32),
  .STORAGE_FIFO_DEPTH  (STORAGE_FIFO_DEPTH)


) read_desc_queue
(
    .clk                     (Clk_i),
    .rst_n                   (Rstn_i),
    
    .dma_s_wdata_i           (descriptor_data_r),
    .dma_s_be_i              (be_rddm_desc),
    .dma_s_write_i           (wen_rddm_desc),    
    .ready_i                 (dma_rddm_desc_ready_i),    
    .desc_rdata_o            (dma_rddm_desc_data_o),
    .data_valid_o            (dma_rddm_desc_valid_o),
    .full_o                  (dma_rddm_desc_full_o),
    .fillcount_o             ()


);

intel_pcie_descriptor_storage_fifo #(
  
  .DESC_DATA_WIDTH (DESC_DATA_WIDTH),
  .PFNUM_WIDTH  (PFNUM_WIDTH),
  .VFNUM_WIDTH  (VFNUM_WIDTH),
  .DESC_BE_WIDTH  (DESC_DATA_WIDTH/32),
  .STORAGE_FIFO_DEPTH  (STORAGE_FIFO_DEPTH)

) read_prio_queue
(
    .clk                     (Clk_i),
    .rst_n                   (Rstn_i),
    
    .dma_s_wdata_i           (descriptor_data_r),
    .dma_s_be_i              (be_rddm_prio),
    .dma_s_write_i           (wen_rddm_prio),
    .ready_i                 (dma_rddm_prio_ready_i),    
    .desc_rdata_o            (dma_rddm_prio_data_o),
    .data_valid_o            (dma_rddm_prio_valid_o),
    .full_o                  (dma_rddm_prio_full_o),
    .fillcount_o             ()


);

intel_pcie_dm_controller_sc_fifo #(
  
  .FIFO_DATA_WIDTH (VFNUM_WIDTH+PFNUM_WIDTH+18),
  .FIFO_DEPTH      (8)

) wrdm_status_queue
(
    .clk                     (Clk_i),
    .rst_n                   (Rstn_i),
    
    .wdata_i                 (dma_wrdm_tx_data_i),
    .write_i                 ((|dma_wrdm_tx_data_i[13:12]) & dma_wrdm_tx_valid_i),
    .read_i                  (read_w_status),
    
    .desc_rdata_o            (wrdm_status_queue_rdata),
    .full_o                  (),
    .empty_o                 (wrdm_status_queue_empty),
    .fillcount_o             ()


);


intel_pcie_dm_controller_sc_fifo #(
  
  .FIFO_DATA_WIDTH (VFNUM_WIDTH+PFNUM_WIDTH+18),
  .FIFO_DEPTH      (8)

) rddm_status_queue
(
    .clk                     (Clk_i),
    .rst_n                   (Rstn_i),
    
    .wdata_i                 (dma_rddm_tx_data_i),
    .write_i                 ((|dma_rddm_tx_data_i[13:12]) & dma_rddm_tx_valid_i),
    .read_i                  (read_r_status),
    
    .desc_rdata_o            (rddm_status_queue_rdata),
    .full_o                  (),
    .empty_o                 (rddm_status_queue_empty),
    .fillcount_o             ()

);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////Interrupt Descriptor Writer
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


  always @ (posedge Clk_i) begin
   int_info_wrdm <= dma_wrdm_tx_data_i[13:12];
   int_info_rddm <= dma_rddm_tx_data_i[13:12];
  end
  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      wr_int_info <= '0;
      wr_int_info_rddm <= 0;
      wr_int_info_wrdm <= 0;    
    end
    else begin
      wr_int_info <= (rddm_int_reg[511] & dma_rddm_tx_valid_i & dma_rddm_tx_data_i[13]) | (wrdm_int_reg[511] & dma_wrdm_tx_valid_i & dma_wrdm_tx_data_i[13]);  
      wr_int_info_rddm <= rddm_int_reg[511] & dma_rddm_tx_valid_i & dma_rddm_tx_data_i[13];
      wr_int_info_wrdm <= wrdm_int_reg[511] & dma_wrdm_tx_valid_i & dma_wrdm_tx_data_i[13];
    end
  end  
  
  always @ (posedge Clk_i) begin
    if (wr_int_info_rddm)
      int_storage_reg[int_storage_wr][1:0] <= int_info_rddm;  
    if (wr_int_info_wrdm)
      int_storage_reg[int_storage_wr][3:2] <= int_info_wrdm;  
  end
  
  assign int_storage_info_wire = int_storage_reg[int_storage_rd[4:1]];
  
  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      int_storage_wr <= '0;
      int_storage_rd <= '0;
    end
    else begin
      if (wr_int_info) begin
        int_storage_wr <= int_storage_wr + 1'b1;
      end
      if (read_int_info) begin
        int_storage_rd <= int_storage_rd +  1'b1;
      end
    end
  end

    always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      fill_int_info <= '0;
    end
    else begin
      if (wr_int_info & !read_int_info)
        fill_int_info <= fill_int_info + 2'd2;        
      if (wr_int_info & read_int_info)
        fill_int_info <= fill_int_info + 1'd1;
      if (read_int_info & !wr_int_info)
        fill_int_info <= fill_int_info - 1'b1;
    end
  end

  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      fill_int_less_than4          <= 1'b1;
      empty_int_storage        <= 1'b1;
    end
    else begin
      fill_int_less_than4 <= (fill_int_info < 3'd4);
      empty_int_storage   <= fill_int_less_than4 & (((fill_int_info[1:0] == 1) & read_int_info & !wr_int_info) | (fill_int_info[1:0] == 0));
    end
  end
  
    
  assign read_int_info = !empty_int_storage & !wen_wrdm_prio & !dma_wrdm_prio_full_o;
 
  assign int_app_spec  = 2'b00;
  assign rddm_int_data = {int_desc_id,int_app_spec,1'b0,1'b0,1'b1,18'h00001,rddm_int_reg[63:0],32'h00000000,rddm_int_reg[95:64]};
  assign wrdm_int_data = {int_desc_id,int_app_spec,1'b0,1'b0,1'b1,18'h00001,wrdm_int_reg[63:0],32'h00000000,wrdm_int_reg[95:64]};
  assign int_desc_wire = int_storage_rd[0] ? wrdm_int_data : rddm_int_data;
  
  always_ff @(posedge Clk_i) begin
    if (!Rstn_i) begin
      int_desc_id      <= '0;
    end
    else begin
      if (read_int_info)
        int_desc_id <= int_desc_id + 1'b1;  
    end
  end

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//msix table starts at offset 0x0
	//pba starts at offset 0x200
	
	localparam  MSIX_NUM   = 128; 
	localparam  MSIX_TABLE_ADDR_W = 15; 
	localparam  TABLE_IDL = 3'b001, TABLE_RD = 3'b010, TABLE_WR = 3'b100;


	logic   [6:0]   table_read_addr;  
	logic   [31:0]  table_read_data; 
	logic 			table_readdatavalid_tmp;
    logic   [2:0]   msix_table_cs, msix_table_ns;
    logic           table_read_en, table_write_en;
	

	always @(posedge Clk_i)
		begin
			if(~Rstn_i)
			   msix_table_cs <= TABLE_IDL;
			else
			   msix_table_cs <= msix_table_ns;
		end
	
	always @*
		case (msix_table_cs)
			TABLE_IDL:
				if (msix_table_read_i)
					msix_table_ns = TABLE_RD;
				else if (msix_table_write_i)
					msix_table_ns = TABLE_WR;
				else 
					msix_table_ns = TABLE_IDL;
			TABLE_RD, TABLE_WR: 
				msix_table_ns = TABLE_IDL;
			
			default: 
				msix_table_cs = TABLE_IDL; 
		
		endcase
	
	assign table_read_en = (msix_table_cs == TABLE_RD); 
	assign table_write_en = (msix_table_cs == TABLE_WR);
	assign msix_table_waitrequest_o = ~(table_read_en | table_write_en); 
	
	always @ (posedge Clk_i)
		begin
			if (~Rstn_i)
                begin
                    table_readdatavalid_tmp <= 1'b0;
                    msix_table_readdatavalid_o <= 1'b0;
                end
			else 
                begin
                    table_readdatavalid_tmp <= table_read_en;
                    msix_table_readdatavalid_o <= table_readdatavalid_tmp;
			    end
		end

	//MSIX table: write to port a and read from port b	

      altera_syncram  msix_table
      (
        .clock0         ( Clk_i),
        .wren_a         ( table_write_en),
        .address_a      ( msix_table_address_i),
        .data_a         ( msix_table_writedata_i            ),
        .rden_b         ( table_read_en                             ),
        .address_b      ( msix_table_address_i     ),
        .q_b            ( msix_table_readdata_o         ),
        .byteena_a      ( msix_table_byteenable_i),
        .byteena_b      ( msix_table_byteenable_i),
        .aclr0          ( 1'b0                 ),
        .aclr1          ( 1'b0                 ),
        .sclr           ( 1'b0                 ), // Not used.
        .addressstall_a ( 1'b0                 ),
        .addressstall_b ( 1'b0                 ),
        .clock1         ( 1'b1                 ),
        .clocken0       ( 1'b1                 ),
        .clocken1       ( 1'b1                 ),
        .clocken2       ( 1'b1                 ),
        .clocken3       ( 1'b1                 ),
        .data_b         ( {32 {1'b1}}          ), // Not used.
        .eccencbypass   ( 1'b0                 ), // Not used.
        .eccencparity   ( 8'b0                 ), // Not used.
        .q_a            (                      ), // Not used.
        .rden_a         ( 1'b1                 ), // Not used.
        .wren_b         ( 1'b0                 ), // Not used.
        .eccstatus      (                      )  // Not used.
        );
      defparam
       msix_table.address_aclr_b              = "NONE",
       msix_table.address_reg_b               = "CLOCK0",
       msix_table.clock_enable_input_a        = "BYPASS",
       msix_table.clock_enable_input_b        = "BYPASS",
       msix_table.clock_enable_output_b       = "BYPASS",
       msix_table.enable_ecc                  = "FALSE",
       msix_table.ecc_pipeline_stage_enabled  = "FALSE",
       msix_table.intended_device_family      = "Stratix 10",
       msix_table.lpm_type                    = "altera_syncram",
       msix_table.maximum_depth               = 512,
       msix_table.numwords_a                  = MSIX_NUM*64,
       msix_table.numwords_b                  = MSIX_NUM*64,
       msix_table.operation_mode              = "DUAL_PORT",
       msix_table.outdata_aclr_b              = "NONE",
       msix_table.outdata_reg_b               = "CLOCK0",
       msix_table.power_up_uninitialized      = "FALSE",
       msix_table.ram_block_type              = "M20K",
       msix_table.widthad_a                   =  13,
       msix_table.widthad_b                   =  13,
       msix_table.width_a                     = 512,
       msix_table.width_b                     = 512,
       msix_table.width_byteena_a             = 64,
       msix_table.width_byteena_b           = 64;


endmodule
