// PCIE_FUNDAMENTAL.cpp : Defines the entry point for the console application.
//
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "PCIE.h"

#define DEMO_PCIE_USER_BAR			PCIE_BAR4
#define DEMO_PCIE_IO_LED_ADDR		0x800000
#define DEMO_PCIE_IO_BUTTON_ADDR	0x800040
#define DEMO_PCIE_MEM_ADDR			0x100000

#define MEM_SIZE					(512*1024) //512KB
typedef enum {
	MENU_LED = 0,
	MENU_BUTTON,
	MENU_DMA_MEMORY,
	MENU_QUIT = 99
} MENU_ID;

void UI_ShowMenu(void)
{
	printf("==============================\n");
	printf("[%d]: Led control\n", MENU_LED);
	printf("[%d]: Button Status Read\n", MENU_BUTTON);
	printf("[%d]: DMA Memory Test\n", MENU_DMA_MEMORY);
	printf("[%d]: Quit\n", MENU_QUIT);
	printf("Plesae input your selection:");
}

int UI_UserSelect(void)
{
	int nSel;
	scanf("%d", &nSel);
	return nSel;
}

bool TEST_LED(PCIE_HANDLE hPCIe)
{
	bool bPass;
	int Mask;

	printf("Please input led conrol mask:");
	scanf("%d", &Mask);

	bPass = PCIE_Write32(hPCIe, DEMO_PCIE_USER_BAR, DEMO_PCIE_IO_LED_ADDR,
		(uint32_t) Mask);
	if (bPass)
		printf("Led control success, mask=%xh\n", Mask);
	else
		printf("Led conrol failed\n");

	return bPass;
}

bool TEST_BUTTON(PCIE_HANDLE hPCIe)
{
	bool bPass = true;
	uint32_t Status;

	bPass = PCIE_Read32(hPCIe, DEMO_PCIE_USER_BAR, DEMO_PCIE_IO_BUTTON_ADDR,
		&Status);
	if (bPass)
		printf("Button status mask:=%xh\n", Status);
	else
		printf("Failed to read button status\n");

	return bPass;
}

char PAT_GEN(uint32_t nIndex)
{
	char Data;
	Data = nIndex & 0xFF;
	return Data;
}

bool TEST_DMA_MEMORY(PCIE_HANDLE hPCIe)
{
	bool bPass = true;
	size_t i;
	const int nTestSize = MEM_SIZE;
	const PCIE_LOCAL_ADDRESS LocalAddr = DEMO_PCIE_MEM_ADDR;
	char *pWrite;
	char *pRead;
	char szError[256];

	pWrite = (char *) malloc(nTestSize);
	pRead = (char *) malloc(nTestSize);
	if (!pWrite || !pRead) {
		bPass = false;
		sprintf(szError, "DMA Memory:malloc failed\n");
	}

	// init test pattern
	for (i = 0; i < nTestSize && bPass; i++)
		*(pWrite + i) = PAT_GEN(i);

	// write test pattern
	if (bPass) {
		bPass = PCIE_DmaWrite(hPCIe, LocalAddr, pWrite, nTestSize);
		if (!bPass)
			sprintf(szError, "DMA Memory:PCIE_DmaWrite failed\n");
	}

	// read back test pattern and verify
	if (bPass) {
		bPass = PCIE_DmaRead(hPCIe, LocalAddr, pRead, nTestSize);

		if (!bPass) {
			sprintf(szError, "DMA Memory:PCIE_DmaRead failed\n");
		} else {
			for (i = 0; i < nTestSize && bPass; i++) {
				if (*(pRead + i) != PAT_GEN(i)) {
					bPass = false;
					sprintf(szError,
						"DMA Memory:Read-back verify unmatch, index = %d, read=%xh, expected=%xh\n",
						i, *(pRead + i), PAT_GEN(i));
				}
			}
		}
	}

	if (!bPass)
		printf("%s", szError);
	else
		printf("DMA-Memory (Size = %d byes) pass\n", nTestSize);

	// free resource
	if (pWrite)
		free(pWrite);
	if (pRead)
		free(pRead);

	return bPass;
}

int main(int argc, char* argv[])
{
	void *lib_handle;
	PCIE_HANDLE hPCIE;
	bool bQuit = false;
	int nSel;

	printf("== Terasic: PCIe Demo Program ==\n");

	lib_handle = PCIE_Load();
	if (!lib_handle) {
		printf("PCIE_Load failed!\n");
		return 0;
	}

	hPCIE = PCIE_Open(DEFAULT_PCIE_VID, DEFAULT_PCIE_DID, 0);
	if (!hPCIE) {
		printf("PCIE_Open failed\n");
	} else {
		while (!bQuit) {
			UI_ShowMenu();
			nSel = UI_UserSelect();
			switch (nSel) {
			case MENU_LED:
				TEST_LED(hPCIE);
				break;
			case MENU_BUTTON:
				TEST_BUTTON(hPCIE);
				break;
			case MENU_DMA_MEMORY:
				TEST_DMA_MEMORY(hPCIE);
				break;
			case MENU_QUIT:
				bQuit = true;
				printf("Bye!\n");
				break;
			default:
				printf("Invalid selection\n");
			} // switch

		} // while

		PCIE_Close(hPCIE);

	}

	PCIE_Unload(lib_handle);
	return 0;

}

